{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE RecordWildCards   #-}

module Pipes.IRC.Message.Render
  ( renderIrcMessage
  , renderIrcMessageNoNL
  , renderNickName )
where

import           Data.ByteString                    (ByteString)
import qualified Data.ByteString.Char8              as C8
import           Data.ByteString.Lazy               (toStrict)
import           Data.ByteString.Lazy.Builder
import           Data.ByteString.Lazy.Builder.ASCII (intDec)
import           Data.Monoid
import           Pipes.IRC.Message.Types

renderIrcMessage :: IrcMessage -> C8.ByteString
renderIrcMessage = toStrict . toLazyByteString . buildIrcMessage True

renderIrcMessageNoNL :: IrcMessage -> ByteString
renderIrcMessageNoNL = toStrict . toLazyByteString . buildIrcMessage False

renderNickName :: NickName -> C8.ByteString
renderNickName = toStrict . toLazyByteString . buildNickName

buildIrcMessage :: Bool -> IrcMessage -> Builder
buildIrcMessage nl IrcMessage {..} =
     buildMsgPrefix  prefix
  <> buildMsgCommand command
  <> buildIrcParams  params
  <> if nl then byteString "\r\n" else mempty

buildMsgPrefix :: Maybe MsgPrefix -> Builder
buildMsgPrefix Nothing = mempty
buildMsgPrefix mp =
     char8 ':'
  <> case mp of
       Just (Left sn)  -> buildServerName sn
       Just (Right nn) -> buildNickName nn
       _               -> mempty
  <> char8 ' '

buildServerName :: ServerName -> Builder
buildServerName = byteString

buildNickName :: NickName -> Builder
buildNickName NickName {..} =
     byteString nick
  <> buildUser user
  <> buildHost host

buildUser :: Maybe ByteString -> Builder
buildUser user =
  case user of
    Nothing   -> mempty
    Just name -> byteString "!" <> byteString name

buildHost :: Maybe ByteString -> Builder
buildHost host =
  case host of
    Nothing   -> mempty
    Just name -> byteString "@" <> byteString name

buildMsgCommand :: MsgCommand -> Builder
buildMsgCommand cmd =
  case cmd of
    Left ircCmd    -> buildIrcCommand ircCmd
    Right ircReply -> buildIrcReply ircReply
  <> byteString " "

buildIrcCommand :: IrcCommand -> Builder
buildIrcCommand cmd =
  case cmd of
    (Unknown name) -> byteString name
    _              -> byteString . C8.pack . show $ cmd

buildIrcReply :: IrcReply -> Builder
buildIrcReply IrcReply {..} =
  let
    h = replyCode `quot` 100
    t = (replyCode - (h * 100)) `quot` 10
    o = replyCode - (h * 100) - (t * 10)
  in intDec h <> intDec t <> intDec o

buildIrcParams :: [IrcParam] -> Builder
buildIrcParams []     = mempty
buildIrcParams (p:ps) =
  buildParam p <> mconcat [char8 ' ' <> buildParam p' | p' <- ps ]

buildParam :: IrcParam -> Builder
buildParam = byteString
