-- | This module contains the pure functions for transforming the state of
-- the IRC server record ('IrcServer') as commands are processed. It  depends
-- on the "User" and "Channel" modules for some transformations of user and
-- channel ('IrcUser' and 'IrcChannel') structures.
module Pipes.IRC.Server.Server
  ( ircAddUser
  , ircDelUser
  , ircHasUser
  , ircHasChan
  , ircJoin
  , ircPart
  , ircInvite
  , ircInviteCheck
  , ircPassCheck
  , ircChangeNick
  , NickKey
  , IrcUser(..)
  , IrcUserMode(..)
  )
where

import           Control.Lens
import qualified Data.Map                 as M
import           Data.Maybe
import qualified Data.Set                 as S
import           Pipes.IRC.Server.Channel
import           Pipes.IRC.Server.Types
import           Pipes.IRC.Server.User
import           Pipes.IRC.Server.Util

-- | Add the user to the server with the given nickname. The
-- nickname should already be in the server's nick list; this
-- function will not check for that condition.
ircAddUser :: NickKey   -- ^ nickname of the user
           -> IrcUser   -- ^ user to add to the server
           -> IrcServer -- ^ server to add the user to
           -> IrcServer -- ^ new server with user added
ircAddUser nn usr = ircUsers %~ M.insert nn usr

-- | Delete the user known by the given nickname from the
-- server. This removes the nickname from the nick list and also
-- from any channels the user is in or invited to.
ircDelUser :: NickKey   -- ^ nickname of the user to delete
           -> IrcServer -- ^ server to delete the user from
           -> IrcServer -- ^ new server with user deleted
ircDelUser nn srv =
  srv & ircNicks %~ sans nn
      & if isn't _Just (srv ^. ircUsers . at nn) then id else
        let
           usr = srv ^. ircUsers . at nn
           uchans = usr ^. traverse . userChannels . to S.elems
           ichans = usr ^. traverse . userInvites . to S.elems
        in
           (ircUsers    %~ sans nn)
         . (ircChannels %~ alterAtKeys (ircPartChan nn) uchans)
         . (ircChannels %~ adjustAtKeys (chanDelInvite nn) ichans)

-- | Check whether a user with the given nickname is known by the
-- server. This only checks for fully-registered users; the nick
-- may still be reserved even if this returns 'False'.
ircHasUser :: NickKey   -- ^ nickname of the user to check for
           -> IrcServer -- ^ server to check
           -> Bool      -- ^ is the user registered?
ircHasUser nn srv = M.member nn (srv ^. ircUsers)

-- | Check whether a channel with the given name is known by the
-- server.
ircHasChan :: ChanKey   -- ^ name of the channel to check for
           -> IrcServer -- ^ server to check
           -> Bool      -- ^ does the channel exist?
ircHasChan cn srv = M.member cn (srv ^. ircChannels)

-- | Add the user with the given nickname to the named channel. If the
-- channel does not exist yet, it will be created. This does not do any
-- permissions checks, it just adds the user.
ircJoin :: NickKey   -- ^ nickname of the joining user
        -> ChanKey   -- ^ name of the channel to join
        -> IrcServer -- ^ server to perform the join on
        -> IrcServer -- ^ new server with join completed
ircJoin uname cn = (ircChannels %~ M.alter alterChan cn)
                 . (ircUsers    %~ M.adjust (userAddChan cn) uname)
  where
    alterChan mChan = Just $
       chanAddUser uname (fromMaybe (newChannel uname) mChan)

-- Helper 'alter' function for ircPart, not exported
ircPartChan :: NickKey -> Maybe IrcChannel -> Maybe IrcChannel
ircPartChan uname (Just chan) =
  case chanDelUser uname chan of
    IrcChannel{ _chanUsers = us }
      | us == S.empty             -> Nothing
    chan'                         -> Just chan'
ircPartChan _     Nothing     = Nothing

-- | Remove the user with the given nickname from the named
-- channel. If this causes the channel to become empty, it will be
-- removed from the server.
ircPart :: NickKey   -- ^ nickname of parting user
        -> ChanKey   -- ^ name of the channel to part from
        -> IrcServer -- ^ server to perform the part on
        -> IrcServer -- ^ new server with part completed
ircPart uname cn srv =
  srv & (ircChannels %~ (M.alter $ ircPartChan uname) cn)
      & (ircUsers    %~ adjustAtKeys (userDelInvite cn) iusers)
  where
    chan = srv ^. ircChannels . at cn
    iusers = chan ^. traverse . chanInvites . to S.elems

-- | Add the user with the given nickname to the invited list for the
-- channel, and add the channel to the user's invited list.
ircInvite :: NickKey   -- ^ nickname of user to invite
          -> ChanKey   -- ^ name of channel user is invited to
          -> IrcServer -- ^ server to perform the invitation on
          -> IrcServer -- ^ new server with invite completed
ircInvite uname cn = (ircChannels %~ M.adjust (chanAddInvite uname) cn)
                   . (ircUsers    %~ M.adjust (userAddInvite cn) uname)

-- | Determine whether the user with the given nickname is disallowed
-- from joining the channel due to the 'InviteOnly' flag and lack of
-- invitation. A 'True' value indicates that the user may join.
ircInviteCheck :: NickKey    -- ^ nickname of possibly invited user
               -> IrcChannel -- ^ name of channel to check
               -> Bool       -- ^ may the user join?
ircInviteCheck n chan =
  not (chanHasModeFlag InviteOnly chan) || chanUserIsInvited n chan

-- | Determine whether the given channel will disallow joining due to
-- a missing or incorrect password. A 'True' value indicates that
-- conditions for joining are met.
ircPassCheck :: Maybe PassKey    -- ^ password supplied by user
             -> IrcChannel       -- ^ channel to check
             -> Bool             -- ^ may the user join?
ircPassCheck k chan =
  not (chanHasPass chan) || (isJust k && chanCheckPass (fromJust k) chan)

-- | Change the nickname of a user from 'old' to 'new', updating the
-- necessary 'IrcServer' structures. No nick collision check is
-- performed; that should be done before calling this.
ircChangeNick :: NickKey   -- ^ old nickname of the user
              -> NickKey   -- ^ new nickname of the user
              -> IrcServer -- ^ server to change the nickname on
              -> IrcServer -- ^ new server with change performed
ircChangeNick old new srv =
  if M.notMember old (srv ^. ircUsers) then
     srv
  else let
     usr = srv ^. ircUsers . at old
     chs = usr ^. traverse . userChannels . to S.elems
  in
     srv & (ircNicks    %~ sans old . S.insert new)
         & (ircUsers    %~ sans old . (at new .~ usr))
         & (ircChannels %~ adjustAtKeys (chanChangeNick old new) chs)
