{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE TemplateHaskell            #-}

module Pipes.IRC.Server.Types
  ( module Pipes.IRC.Server.Types
  , HostPreference (..)
  , ServiceName
  , SockAddr
  , Socket
  )
where

import           Control.Concurrent.STM  (TVar)
import           Control.Lens
import           Control.Monad.RWS       (RWS)
import           Data.ByteString         (ByteString)
import           Data.Map                (Map)
import           Data.Set                (Set)
import           Data.Time.Clock         (UTCTime)
import           Pipes.Concurrent        (Output)
import           Pipes.IRC.Message.Types (IrcMessage, NickName)
import           Pipes.Network.TCP       (HostPreference (..), ServiceName,
                                          SockAddr, Socket)

-- * Data structures for managing the pure component of the server

-- | A type alias for nicknames
type NickKey = ByteString

-- | A type alias for channel names
type ChanKey = ByteString

-- | A type alias for password keys
type PassKey = ByteString

-- | A sequence of 'IrcEvent's
type IrcEvents = [IrcEvent]

-- | An 'IrcEvent' represents an action that should occur due to the processing
-- of actions within the 'IrcMonad'. It includes sending messages, closing
-- connections, etc.
data IrcEvent = Msg { _outMsg  :: !IrcMessage
                      -- ^ An 'IrcMessage' to send over the network
                    , _outDest :: ![Int]
                      -- ^ The connection ids to send the message on
                    }
              | Close { _closeConn :: Int
                        -- ^ The connection id to close
                      }
              | Pong { _pongConn :: Int
                       -- ^ The connection id to send a PONG message to
                     }
               deriving (Show)
makeLenses ''IrcEvent

-- | An 'IrcUserMode' represents some privilege or state of a user. They are
-- used for various different purposes.
data IrcUserMode = Away          -- ^ the user is marked as away from keyboard
                 | Invisible     -- ^ the user does not show in WHO listings
                 | WallOps       -- ^ the user receives WALLOPS messages
                 | Restricted    -- ^ ???
                 | Oper          -- ^ the user is an IRC operator
                 | LocalOper     -- ^ the user is a local server operator
                 | ServerNotices -- ^ the user receives server notices
            deriving (Show, Eq, Enum, Ord, Bounded)

-- | An 'IrcUser' record tracks the state of a registered user that is not
-- channel-specific.
data IrcUser =
  IrcUser { _userServerName :: !ByteString
            -- ^ the server that the user connected from
          , _userModes      :: !(Set IrcUserMode)
            -- ^ the mode flags that the user has set
          , _userChannels   :: !(Set ChanKey)
            -- ^ channels that the user is a current member of
          , _userConn       :: !Int
            -- ^ an integer identifying the user's connection
          , _userInvites    :: !(Set ChanKey)
            -- ^ the set of channels that the user is invited to
          } deriving (Show, Eq)
makeLenses ''IrcUser

-- | The 'IrcChanModeFlags' represent some state that is specific to a channel.
-- They are used for various purposes.
data IrcChanModeFlags = Anonymous      -- ^ all communication is anonymized
                      | InviteOnly     -- ^ only invited users may join
                      | Moderated      -- ^ only ops or voice may speak
                      | NoOutsideMsgs  -- ^ only members may speak
                      | Quiet          -- ^ ???
                      | Private        -- ^ channel shows up as private in list
                      | Secret         -- ^ channel does not appear in list
                      | TopicOperOnly  -- ^ topic may only be set by ops
            deriving (Show, Eq, Enum, Ord, Bounded)

-- | An 'IrcChannel' record tracks the state of a channel.
data IrcChannel =
  IrcChannel { _chanTopic     :: !(Maybe ByteString)
               -- ^ the topic of discussion for the channel
             , _chanPass      :: !(Maybe ByteString)
               -- ^ key that must be given to join the channel
             , _chanModeFlags :: !(Set IrcChanModeFlags)
               -- ^ various configuration options for channel behavior
             , _chanUsers     :: !(Set NickKey)
               -- ^ users that have joined the channel
             , _chanOpers     :: !(Set NickKey)
               -- ^ users in the channel with ops
             , _chanVoices    :: !(Set NickKey)
               -- ^ users in the channel with voice
             , _chanInvites   :: !(Set NickKey)
               -- ^ users that have been invited to the channel
             } deriving (Show, Eq)
makeLenses ''IrcChannel

-- | An 'IrcServer' record tracks the pure state of the server.
data IrcServer =
  IrcServer { _ircNicks    :: !(Set NickKey)
              -- ^ the nicknames currently known to be in use
            , _ircUsers    :: !(Map NickKey IrcUser)
              -- ^ users that have been registered
            , _ircChannels :: !(Map ChanKey IrcChannel)
              -- ^ channels that have at least one user
            , _ircVersion  :: !ByteString
              -- ^ version of the server
            } deriving (Show)
makeLenses ''IrcServer

-- | Configuration options for the server
data IrcConfig =
  IrcConfig { _ircPort     :: !ServiceName
              -- ^ port on which the server should listen for connections
            , _ircHost     :: !HostPreference
              -- ^ interface name on which the server should listen
            , _ircHostName :: !ByteString
              -- ^ official server name of the server
            , _ircMotd     :: ![ByteString]
              -- ^ MOTD to send to new connections to the server
            , _ircPass     :: !(Maybe ByteString)
              -- ^ a secret password that must be entered to register
            } deriving (Show)
makeLenses ''IrcConfig

-- | Data that tracks the registration information of clients
data RegState = Unreg   { _rcvdPass :: !(Maybe ByteString)
                          -- ^ the password given via PASS
                        , _rcvdNick :: !(Maybe NickKey)
                          -- ^ the nickname given during registration
                        , _rcvdName :: !(Maybe ByteString)
                          -- ^ the real name given during registration
                        }
              | RegUser { _regdNick :: !NickName
                          -- ^ information about a user from registration
                        }
  deriving (Show)
makeLenses ''RegState

-- * Data structures for managing the IO portion of the server

data IrcConnection =
  IrcConnection { _sock    :: !Socket
                , _addr    :: !SockAddr
                , _hname   :: !(Maybe ByteString)
                , _out     :: !(Output IrcMessage)
                , _reg     :: !RegState
                , _lastCom :: !UTCTime
                , _gotPong :: !Bool
                }
makeLenses ''IrcConnection

data ServerState =
  ServerState { _ircState       :: !(TVar IrcServer)
              , _ircConfig      :: !IrcConfig
              , _ircConnections :: !(TVar (Map Int IrcConnection))
              , _ircConnIds     :: !(TVar Int)
              }
makeLenses ''ServerState

data ClientState =
  ClientState { _clientReg    :: !RegState
              , _clientServer :: !IrcServer
              , _clientHost   :: !(Maybe ByteString)
              , _clientConn   :: !Int
              } deriving (Show)
makeLenses ''ClientState

type IrcMonad = RWS IrcConfig IrcEvents ClientState
